
/*
 * exec.c
 *   Query execution and data retrieval functions.
 *
 * Copyright (c) 2008 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

/*
 * Each param value requires an oid, ptr and 2 ints (oid, value, length, format).
 * This makes the maximum size of a param 20 bytes. A stack size of 4k
 * would allow for 204 query params.	If more params are needed,
 * heap memory is used. Needing more than 204 param values in a query
 * is very rare, although possible.
 */
#define PARAM_STACKSIZE 4096

#define BUILD_ARRAYS(rettype) \
	rettype r; \
	char *buf   = NULL; \
	Oid *oids   = NULL; \
	char **vals = NULL; \
	int *lens   = NULL; \
	int *fmts   = NULL; \
	int vcnt    = 0; \
	char stackbuffer[PARAM_STACKSIZE]; \
\
	if(!conn) \
	{ \
		PQseterror("PGconn cannot be NULL"); \
		return (rettype)(0); \
	} \
\
	if(param) \
	{ \
		buf = stackbuffer; \
		if(!buildArrays(param, &buf, &oids, &vals, &lens, &fmts)) \
			return (rettype)(0); \
		vcnt = param->vcnt; \
	} \
\
	PQseterror(NULL)

#define RETURN_RESULT \
	if(param) \
	{ \
		if(buf && buf != stackbuffer) \
			free(buf); \
	} \
	return r

static int
buildArrays(PGparam *param, char **buf, Oid **oids,
	char ***vals, int **lens, int **fmts);

static PGresult *
copyExecError(PGconn *conn, PGresult *r);


int
PQgetf(const PGresult *res, int tup_num, const char *fieldspec, ...)
{
	int r;
	PGtypeHandler *h;
	PGtypeArgs args;
	int typpos = 0;
	char *s;
	int flags;
	Oid ftype;
	char schema[PQT_MAXIDLEN + 1];
	char typname[PQT_MAXIDLEN + 1];
	char tmp[200];
	PGtypeData *resData;

	if(!res)
	{
		PQseterror("PGresult cannot be NULL");
		return FALSE;
	}

	if(!(resData = PQresultHookData(res, PQT_OBJHOOKSNAME)))
	{
		PQseterror("PGresult at %p has no " PQT_OBJHOOKSNAME " hook data", res);
		return FALSE;
	}

	PQseterror(NULL);
	va_start(args.ap, fieldspec);

	while(fieldspec && *fieldspec)
	{
		int specid = *fieldspec++;

		if(specid != '%' && specid != '#')
			continue;

		typpos++;

		if(!(s = pqt_parsetype(fieldspec, schema, typname, &flags, typpos)))
		{
			va_end(args.ap);
			return FALSE;
		}

		if(flags & TYPFLAG_INVALID)
		{
			va_end(args.ap);
			return FALSE;
		}

		fieldspec = (char *)s;

		/* find the handler */
		if(!(h = pqt_gethandler(
			resData->typhandlers,
			resData->typhcnt,
			schema,
			typname)))
		{
			PQseterror("Uknown type '%s' (position %d)",
				pqt_fqtn(tmp, sizeof(tmp), schema, typname), typpos);
			va_end(args.ap);
			return FALSE;
		}

		if(!h->typget)
		{
			h = pqt_gethandlerbyid(
				resData->typhandlers, resData->typhcnt, h->base_id);

			if(!h || !h->typget)
			{
				PQseterror(
					"Type '%s' doesn't support put operations (position %d)",
					pqt_fqtn(tmp, sizeof(tmp), h->typschema, h->typname), typpos);
				va_end(args.ap);
				return FALSE;
			}
		}

		if(specid == '#')
			args.get.field_num = PQfnumber(res, va_arg(args.ap, const char *));
		else
			args.get.field_num = va_arg(args.ap, int);

		/* simplify life for handlers by checking getvalue's return here. */
		if(args.get.field_num < 0 || !PQgetvalue(res, tup_num, args.get.field_num))
		{
			PQseterror(
				"Invalid tup_num[%d].field_num[%d] (position %d)",
				tup_num, args.get.field_num, typpos);
			va_end(args.ap);
			return FALSE;
		}

		ftype = PQftype(res, args.get.field_num);
		if(((flags & TYPFLAG_ARRAY) && ftype != h->typoid_array) ||
			 (!(flags & TYPFLAG_ARRAY) && ftype != h->typoid))
		{
			Oid oid = (flags & TYPFLAG_ARRAY) ? h->typoid_array : h->typoid;
			PQseterror(
				"Trying to get type %u '%s' but server returned %u (position %d)",
				oid, pqt_fqtn(tmp, sizeof(tmp), h->typschema, h->typname),
				ftype, typpos);
			va_end(args.ap);
			return FALSE;
		}

		args.is_put       = 0;
		args.get.result   = (PGresult *)res;
		args.format       = PQfformat(res, args.get.field_num);
		args.fmtinfo      = &resData->fmtinfo;
		args.get.tup_num  = tup_num;
		args.is_ptr       = (flags & TYPFLAG_POINTER) ? 1 : 0;
		args.typpos       = typpos;
		args.typhandler   = h;
		args.errorf       = pqt_argserrorf;
		args.super        = pqt_argssuper;

		if(flags & TYPFLAG_ARRAY)
			r = pqt_get_array(&args);
		else
			r = h->typget(&args);

		if(r == -1)
		{
			va_end(args.ap);
			return FALSE;
		}
	}

	va_end(args.ap);
	return TRUE;
}

/* --------------------------------
 * Exec and Send functions
 */

PGresult *
PQparamExec(PGconn *conn, PGparam *param, const char *command,
	int resultFormat)
{
	BUILD_ARRAYS(PGresult *);

	r = PQexecParams(
		conn,
		command,
		vcnt,
		oids,
		(const char *const * )vals,
		lens,
		fmts,
		resultFormat);

	r = copyExecError(conn, r);

	RETURN_RESULT;
}

int
PQparamSendQuery(PGconn *conn, PGparam *param, const char *command,
	int resultFormat)
{
	BUILD_ARRAYS(int);

	r = PQsendQueryParams(
		conn,
		command,
		vcnt,
		oids,
		(const char *const * )vals,
		lens,
		fmts,
		resultFormat);

	if(!r)
		PQseterror("PGconn: %s", PQerrorMessage(conn));

	RETURN_RESULT;
}

PGresult *
PQparamExecPrepared(PGconn *conn, PGparam *param, const char *stmtName,
	int resultFormat)
{
	BUILD_ARRAYS(PGresult *);

	r = PQexecPrepared(
		conn,
		stmtName,
		vcnt,
		(const char *const * )vals,
		lens,
		fmts,
		resultFormat);

	r = copyExecError(conn, r);

	RETURN_RESULT;
}

int
PQparamSendQueryPrepared(PGconn *conn, PGparam *param, const char *stmtName,
	int resultFormat)
{
	BUILD_ARRAYS(int);

	r = PQsendQueryPrepared(
		conn,
		stmtName,
		vcnt,
		(const char *const * )vals,
		lens,
		fmts,
		resultFormat);

	if(!r)
		PQseterror("PGconn: %s", PQerrorMessage(conn));

	RETURN_RESULT;
}

/*
 * Assigns param values to param arrays, for use with postgres
 * parameter API. 'buf' is expected to be PARAM_STACKSIZE bytes.  If more
 * memory is required, memory is allocated and assigned to *buf, which
 * must be freed by caller.  To determine if *buf was allocated, compare
 * its address to the initially provided stack address.
 * Returns 1 on success and 0 on error.
 */

static int
buildArrays(PGparam *param, char **buf, Oid **oids,
	char ***vals, int **lens, int **fmts)
{
	int n;

	/* no params to assign */
	if(param->vcnt == 0)
		return 1;

	/* required memory size for the 4 param arrays */
	n = (int)((sizeof(void *) * param->vcnt) + /* values */
			((sizeof(int) * 2) * param->vcnt) +	   /* lengths and formats */
			(sizeof(Oid) * param->vcnt));				   /* oids */

	/* required memory is too large for stack buffer, get some heap */
	if(n > PARAM_STACKSIZE)
	{
		char *p;

		if(!(p = (char *)malloc(n)))
		{
			PQseterror(PQT_OUTOFMEMORY);
			return 0;
		}

		*buf = p;
	}

	/* give arrays memory from buffer, which could be stack or heap. */
	*vals = (char **)*buf;
	*lens = (int *)(*buf + (sizeof(void *) * param->vcnt));
	*fmts = (*lens) + param->vcnt;
	*oids = (Oid *)((*fmts) + param->vcnt);

	/* loop param values and assign value, length, format and oid to arrays. */
	for(n=0; n < param->vcnt; n++)
	{
		(*oids)[n] = param->vals[n].oid;
		(*vals)[n] = param->vals[n].data;
		(*lens)[n] = param->vals[n].datal;
		(*fmts)[n] = param->vals[n].format;
	}

	return 1;
}

static PGresult *
copyExecError(PGconn *conn, PGresult *r)
{
	if(!r)
	{
		PQseterror("PGconn: %s", PQerrorMessage(conn));
		return NULL;
	}

	switch(PQresultStatus(r))
	{
		case PGRES_COMMAND_OK:
		case PGRES_TUPLES_OK:
		case PGRES_EMPTY_QUERY:
			break;

		default:
			PQseterror("PGresult: %s", PQresultErrorMessage(r));
			PQclear(r);
			r = NULL;
			break;
	}

	return r;
}

