
/*
 * handler.c
 *   Type handler management functions, the core of the handler system.
 *
 * Copyright (c) 2008 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

/* -------------------------
 * DEFAULT HANDLERS: builtin postgresql types
 */

#define DEFAULT_HANDLER_COUNT \
	(int)(sizeof(pg_handlers) / sizeof(pg_handlers[0]))

/* Built-in handler defaults:
 * base_id, nattrs, attrsize, attrs, orig_typname, orig_typput, orig_typget
 */
#define __HANDLER_DEFAULTS__ -1, 0, NULL, NULL, NULL, NULL

/* id, schema, typname, typlen, oid, arroid, typput, typget, defaults */
static PGtypeHandler pg_handlers[] = {
	/* character types */
	{ 0,  "pg_catalog",  "char",            1,  CHAROID,         1002,  pqt_put_char,         pqt_get_char,        __HANDLER_DEFAULTS__},
	{ 1,  "pg_catalog",  "varchar",        -1,  VARCHAROID,      1015,  pqt_put_text,         pqt_get_text,        __HANDLER_DEFAULTS__}, /* supports ptr */
	{ 2,  "pg_catalog",  "bpchar",         -1,  BPCHAROID,       1014,  pqt_put_text,         pqt_get_text,        __HANDLER_DEFAULTS__}, /* supports ptr */
	{ 3,  "pg_catalog",  "text",           -1,  TEXTOID,         1009,  pqt_put_text,         pqt_get_text,        __HANDLER_DEFAULTS__}, /* supports ptr */

	/* boolean types */
	{ 4,  "pg_catalog",  "bool",            1,  BOOLOID,         1000,  pqt_put_bool,         pqt_get_bool,        __HANDLER_DEFAULTS__},

	/* numeric types */
	{ 5,  "pg_catalog",  "int2",            2,  INT2OID,         1005,  pqt_put_int2,         pqt_get_int2,        __HANDLER_DEFAULTS__},
	{ 6,  "pg_catalog",  "int4",            4,  INT4OID,         1007,  pqt_put_int4,         pqt_get_int4,        __HANDLER_DEFAULTS__},
	{ 7,  "pg_catalog",  "int8",            8,  INT8OID,         1016,  pqt_put_int8,         pqt_get_int8,        __HANDLER_DEFAULTS__},
	{ 8,  "pg_catalog",  "float4",          4,  FLOAT4OID,       1021,  pqt_put_float4,       pqt_get_float4,      __HANDLER_DEFAULTS__},
	{ 9,  "pg_catalog",  "float8",          8,  FLOAT8OID,       1022,  pqt_put_float8,       pqt_get_float8,      __HANDLER_DEFAULTS__},
	{10,  "pg_catalog",  "numeric",        -1,  NUMERICOID,      1231,  pqt_put_numeric,      pqt_get_numeric,     __HANDLER_DEFAULTS__},

	/* bytea types */
	{11,  "pg_catalog",  "bytea",          -1,  BYTEAOID,        1001,  pqt_put_bytea,        pqt_get_bytea,       __HANDLER_DEFAULTS__}, /* supports ptr */

	/* geometric types */
	{12,  "pg_catalog",  "point",          16,  POINTOID,        1017,  pqt_put_point,        pqt_get_point,       __HANDLER_DEFAULTS__},
	{13,  "pg_catalog",  "lseg",           32,  LSEGOID,         1018,  pqt_put_lseg,         pqt_get_lseg,        __HANDLER_DEFAULTS__},
	{14,  "pg_catalog",  "box",            32,  BOXOID,          1020,  pqt_put_box,          pqt_get_box,         __HANDLER_DEFAULTS__},
	{15,  "pg_catalog",  "circle",         24,  CIRCLEOID,        719,  pqt_put_circle,       pqt_get_circle,      __HANDLER_DEFAULTS__},
	{16,  "pg_catalog",  "path",           -1,  PATHOID,         1019,  pqt_put_path,         pqt_get_path,        __HANDLER_DEFAULTS__},
	{17,  "pg_catalog",  "polygon",        -1,  POLYGONOID,      1027,  pqt_put_polygon,      pqt_get_polygon,     __HANDLER_DEFAULTS__},

	/* monetary types */
	{18,  "pg_catalog",  "money",           8,  CASHOID,          791,  pqt_put_money,        pqt_get_money,       __HANDLER_DEFAULTS__},

	/* network address typess */
	{19,  "pg_catalog",  "inet",           -1,  INETOID,         1041,  pqt_put_inet,         pqt_get_inet,        __HANDLER_DEFAULTS__},
	{20,  "pg_catalog",  "cidr",           -1,  CIDROID,          651,  pqt_put_inet,         pqt_get_cidr,        __HANDLER_DEFAULTS__},
	{21,  "pg_catalog",  "macaddr",         6,  MACADDROID,      1040,  pqt_put_macaddr,      pqt_get_macaddr,     __HANDLER_DEFAULTS__},

	/* date & time types */
	{22,  "pg_catalog",  "time",            8,  TIMEOID,         1183,  pqt_put_time,         pqt_get_time,        __HANDLER_DEFAULTS__},
	{23,  "pg_catalog",  "timetz",         12,  TIMETZOID,       1270,  pqt_put_timetz,       pqt_get_timetz,      __HANDLER_DEFAULTS__},
	{24,  "pg_catalog",  "date",            4,  DATEOID,         1182,  pqt_put_date,         pqt_get_date,        __HANDLER_DEFAULTS__},
	{25,  "pg_catalog",  "timestamp",       8,  TIMESTAMPOID,    1115,  pqt_put_timestamp,    pqt_get_timestamp,   __HANDLER_DEFAULTS__},
	{26,  "pg_catalog",  "timestamptz",     8,  TIMESTAMPTZOID,  1185,  pqt_put_timestamptz,  pqt_get_timestamptz, __HANDLER_DEFAULTS__},
	{27,  "pg_catalog",  "interval",       16,  INTERVALOID,     1187,  pqt_put_interval,     pqt_get_interval,    __HANDLER_DEFAULTS__},

	/* object identifier types */
	{28,  "pg_catalog",  "oid",             4,  OIDOID,          1028,  pqt_put_int4,         pqt_get_int4,        __HANDLER_DEFAULTS__},

	/* UUID types */
	{29,  "pg_catalog",  "uuid",           16,  UUIDOID,         2951,  pqt_put_uuid,         pqt_get_uuid,        __HANDLER_DEFAULTS__},

	/* Record types (composites) */
	{30,  "pg_catalog",  "record",         -1,  RECORDOID,          0,  pqt_put_record,       pqt_get_record,      __HANDLER_DEFAULTS__},

	/* pqt types */
	{31,  "pqt",         "str",            -1,  InvalidOid,         0,  pqt_put_str,          NULL,                __HANDLER_DEFAULTS__}, /* supports ptr */
	{32,  "pqt",         "null",           -1,  InvalidOid,         0,  pqt_put_null,         NULL,                __HANDLER_DEFAULTS__},

	/* more character types */
	{33,  "pg_catalog",  "name",           -1,  NAMEOID,         1003,  pqt_put_text,         pqt_get_text,        __HANDLER_DEFAULTS__}  /* supports ptr */
};

/* -------------------------
 */

static char *
parseId(char *id, char **start, int *len, int *flags, int typpos);

static int
expandHandlers(PGtypeData *typeData);

static int
createHandler(PGconn *conn, PGtypeData *connData, PGtypeHandler *handler,
	char *schema, char *typname, PGtypeProc typput, PGtypeProc typget);

static int
registerSubClass(PGtypeData *connData, const char *basestr,
	char *sub_typschema, char *sub_typname, const char *orig_typname,
	PGtypeProc typput, PGtypeProc typget);

static int
registerRecord(PGconn *conn, PGtypeData *connData,
	char *comp_typschema, char *comp_typname, const char *orig_typname);



int
PQregisterTypeHandler(PGconn *conn, const char *type_name,
	PGtypeProc typput, PGtypeProc typget)
{
	char *s;
	char typname[PQT_MAXIDLEN + 1];
	char typschema[PQT_MAXIDLEN + 1];
	PGtypeHandler *h;
	int flags;
	char fqtn[200];
	PGtypeData *connData;

	PQseterror(NULL);

	if(!conn)
	{
		PQseterror("PGconn cannot be NULL");
		return FALSE;
	}

	if(!type_name)
	{
		PQseterror("type_name cannot be NULL");
		return FALSE;
	}

	if(!(connData = PQhookData(conn, PQT_OBJHOOKSNAME)))
	{
		PQseterror("PGconn is missing hook data");
		return FALSE;
	}

	if(!(s = pqt_parsetype(type_name, typschema, typname, &flags, 1)))
		return 0;

	if(flags & TYPFLAG_INVALID)
		return FALSE;

	if(flags & TYPFLAG_ARRAY)
	{
		PQseterror(
			"Cannot use an array[] during a type handler registration.");
		return FALSE;
	}

	if(flags & TYPFLAG_POINTER)
	{
		PQseterror(
			"Cannot use a type* during a type handler registration.");
		return FALSE;
	}

	/* Sub-class register: "typschema.typname=base_schema.base_name" */
	if(*s == '=')
		return registerSubClass(connData, s + 1, typschema,
			typname, type_name, typput, typget);

	/* composite/record registration */
	if(!typput && !typget)
		return registerRecord(conn, connData, typschema, typname, type_name);

	/* -----------------
	 * BASE TYPE
	 */

	if(pqt_gethandler(connData->typhandlers, connData->typhcnt,
			typschema, typname))
	{
		PQseterror("type '%s' already exists",
			pqt_fqtn(fqtn, sizeof(fqtn), typschema, typname));
		return FALSE;
	}

	if(!expandHandlers(connData))
		return FALSE;

	h = &connData->typhandlers[connData->typhcnt];
	if(createHandler(conn, connData, h, typschema, typname,
			typput, typget) == -1)
		return FALSE;

	h->orig_typname = strdup(type_name);
	h->orig_typput = typput;
	h->orig_typget = typget;
	connData->typhcnt++;
	return TRUE;
}

/* Do not call when hcnt is 0.  This returns NULL when malloc fails.
 * Passing in 0 could only do the same thing (ambiguos), so it is simply
 * not handled and may dump core.
 */
PGtypeHandler *
pqt_duphandlers(PGtypeHandler *handlers, int hcnt)
{
	int i,n;
	PGtypeHandler *h = (PGtypeHandler *)malloc(hcnt * sizeof(PGtypeHandler));

	if(!h)
		return NULL;

	memcpy(h, handlers, hcnt * sizeof(PGtypeHandler));

	/* deep copy schema, typname and the PGrecordAttDesc array */
	for(i=0; i < hcnt; i++)
	{
		if(handlers[i].typschema)
			h[i].typschema = strdup(handlers[i].typschema);

		if(handlers[i].typname)
			h[i].typname = strdup(handlers[i].typname);

		if(handlers[i].orig_typname)
			h[i].orig_typname = strdup(handlers[i].orig_typname);

		if(handlers[i].nattrs == 0)
			continue;

		h[i].attDescs = (PGrecordAttDesc *)malloc(
			handlers[i].nattrs * sizeof(PGrecordAttDesc));
		if(!h[i].attDescs)
		{
			pqt_freehandlers(h, i+1);
			return NULL;
		}

		memcpy(h[i].attDescs, handlers[i].attDescs,
			handlers[i].nattrs * sizeof(PGrecordAttDesc));

		for(n=0; n < handlers[i].nattrs; n++)
			h[i].attDescs[n].attname = strdup(
				handlers[i].attDescs[n].attname);
	}

	return h;
}

void
pqt_freehandlers(PGtypeHandler *handlers, int hcnt)
{
	int i;

	for(i=0; i < hcnt; i++)
	{
		if(handlers[i].typschema)
			free(handlers[i].typschema);

		if(handlers[i].typname)
			free(handlers[i].typname);

		if(handlers[i].orig_typname)
			free(handlers[i].orig_typname);

		if(handlers[i].attDescs)
		{
			int n;
			for(n=0; n < handlers[i].nattrs; n++)
				free(handlers[i].attDescs[n].attname);
			free(handlers[i].attDescs);
		}

		handlers[i].typschema    = NULL;
		handlers[i].typname      = NULL;
		handlers[i].orig_typname = NULL;
		handlers[i].nattrs       = 0;
		handlers[i].attDescs     = NULL;
	}

	if(handlers)
		free(handlers);
}

PGtypeHandler *
pqt_gethandler(PGtypeHandler *handlers, int hcnt,
	const char *schema, const char *typname)
{
	int i;
	int noschema = !schema || !*schema;

	if(!typname || !*typname)
		return NULL;

	/* user registered types are searched first */
	for(i=0; i < hcnt; i++)
	{
		if((noschema || strcmp(handlers[i].typschema, schema)==0) &&
			 strcmp(handlers[i].typname, typname)==0)
		{
			return &handlers[i];
		}
	}

	/* builtin types searched last */
	for(i=0; i < DEFAULT_HANDLER_COUNT; i++)
	{
		if((noschema || strcmp(pg_handlers[i].typschema, schema)==0) &&
			 strcmp(pg_handlers[i].typname, typname)==0)
		{
			return &pg_handlers[i];
		}
	}

	return NULL;
}

PGtypeHandler *
pqt_gethandlerbyid(PGtypeHandler *handlers, int hcnt, int id)
{
	if(id <= -1)
		return NULL;

	if(id < DEFAULT_HANDLER_COUNT)
		return &pg_handlers[id];

	id -= DEFAULT_HANDLER_COUNT;
	if(id >= hcnt)
		return NULL;

	return &handlers[id];
}

int
pqt_argssuper(PGtypeArgs *args, ...)
{
	int r;
	va_list ap;
	PGtypeHandler *baseclass;
	PGtypeHandler *subclass = args->typhandler;
	PGtypeData *resData = NULL;

	if(!args->is_put)
	{
		resData = PQresultHookData(args->get.result, PQT_OBJHOOKSNAME);
		if(!resData)
			return args->errorf(args, "PGresult is missing hook data");
	}

	/* should always work, but play it safe */
	baseclass = pqt_gethandlerbyid(
		args->is_put ? args->put.param->typhandlers : resData->typhandlers,
		args->is_put ? args->put.param->typhcnt : resData->typhcnt,
		subclass->base_id);

	if(!baseclass)
		return args->errorf(args, "type handler has no base type");

	args->typhandler = baseclass;

	va_copy(ap, args->ap);
	va_start(args->ap, args);
	r = args->is_put ? baseclass->typput(args) : baseclass->typget(args);
	va_copy(args->ap, ap);

	args->typhandler = subclass;
	return r;
}

/* FQTN standards for Fully Qualified Type Name.  Returns a pointer to out.
 * Only returns NULL if out is NULL or outl <= 0.
 */
char *
pqt_fqtn(char *out, size_t outl, const char *schema, const char *typname)
{
	int r;
	int have_schema = schema && *schema;

	if(!out || outl<=0)
		return NULL;

	*out = 0;
	if(!typname || !*typname)
		return out;

	r = pqt_snprintf(out, outl, "%s%s%s",
		have_schema ? schema : "",
		have_schema ? "." : "",
		typname);

	if(r < 0)
	{
		*out = 0;
		return out;
	}

	return out;
}

/* only checks builin types or pqt types.  User registered types must
 * be checked by the user's handler functions.
 */
int
pqt_allowsptr(PGtypeHandler *h)
{
	/* builtin types that support ptr always have a schema name. */
	if(!h->typschema)
		return FALSE;

	/* pg_catalog.[bpchar, varchar, name, text, bytea] */
	if(strcmp(h->typschema, "pg_catalog")==0)
	{
		if(strcmp(h->typname, "bpchar") &&
			 strcmp(h->typname, "varchar") &&
			 strcmp(h->typname, "text") &&
			 strcmp(h->typname, "bytea") &&
			 strcmp(h->typname, "name"))
		{
			return FALSE;
		}
	}
	/* pqt.str */
	else if(strcmp(h->typschema, "pqt")==0 && strcmp(h->typname, "str"))
	{
		return FALSE;
	}

	return TRUE;
}

void
pqt_getfmtinfo(const PGconn *conn, PGtypeFormatInfo *info)
{
	const char *value;

	memset(info, 0, sizeof(PGtypeFormatInfo));

	if((value = PQparameterStatus(conn, "DateStyle")))
		pqt_strcpy(info->datestyle, sizeof(info->datestyle), value);

	if((value = PQparameterStatus(conn, "integer_datetimes")))
		info->integer_datetimes = strcmp(value, "on")==0 ? TRUE : FALSE;

	info->sversion = PQserverVersion(conn);
	info->pversion = PQprotocolVersion(conn);
}

/* Parse a type identifer name (schema qualified or not) from spec. spec
 * must point to the first char after the % sign, which maybe a double quote.
 *
 * spec - pointer to typname, just after the '%' or '#'
 * schema - buffer to receive schema (PQT_MAXIDLEN bytes)
 * typname - buffer to receive typname (PQT_MAXIDLEN bytes)
 * flags - a pointer to an int that is set one or more TYPFLAG_xxx
 * typpos - 1-based position of spec in specifier string (0 for unknown)
 */
char *
pqt_parsetype(const char *spec, char *schema, char *typname,
	int *flags, int typpos)
{
	int i;
	char *start;
	int len=0;
	char *s = (char *)spec;

	if(!(s = parseId(s, &start, &len, flags, typpos)))
		return NULL;

	/* not a valid specifer, false positive like "(x % y) = 0" */
	if(*flags & TYPFLAG_INVALID)
		return s;

	*schema = 0;
	if(*s == '.')
	{
		memcpy(schema, start, len);
		schema[len] = 0;
		if(*flags & TYPFLAG_CASEFOLD)
			for(i=0; i < len; i++)
				schema[i] = pqt_tolower(schema[i]);

		/* now get typname */
		if(!(s = parseId(++s, &start, &len, flags, typpos)))
			return NULL;

		if(*flags & TYPFLAG_INVALID)
			return s;
	}

	memcpy(typname, start, len);
	typname[len] = 0;
	if(*flags & TYPFLAG_CASEFOLD)
		for(i=0; i < len; i++)
			schema[i] = pqt_tolower(schema[i]);

	return s;
}

static char *
parseId(char *id, char **start, int *len, int *flags, int typpos)
{
	char *p = id;

	*flags = 0;
	*start = NULL;
	*len = 0;

	if(*p == '"')
		p++;

	/* check first character */
	if(!isalpha(*p) && *p != '_')
	{
		*flags |= TYPFLAG_INVALID;
		PQseterror(
			"Invalid first character for identifier '%c' (pos:%d)", *p, typpos);
		return p;
	}

	if(*id == '"')
	{
		id++;
		if(!(p = strchr(id, '"')))
		{
			*flags |= TYPFLAG_INVALID;
			PQseterror("Unterminated double quote '%s' (pos:%d)",
				id-1, typpos);
			return p;
		}

		*len = (int)(p - id);
		*start = id;
		p++;
	}
	else
	{
		for(p=id+1; isalnum(*p) || *p=='_'; p++) ;

		*len = (int)(p - id);
		*start = id;
		*flags |= TYPFLAG_CASEFOLD;
	}

	/* range check */
	if(*len == 0 || *len > PQT_MAXIDLEN)
	{
		*flags |= TYPFLAG_INVALID;
		PQseterror("Identifier out of range %d (pos:%d), range is 1 to %d",
			*len, typpos, PQT_MAXIDLEN);
		return p;
	}

	/* direct pointer request */
	if(*p == '*')
	{
		p++;
		*flags |= TYPFLAG_POINTER;
	}

	/* Is this an array?  Ex. %int4[] or %"a b"[] */
	if(p[0] == '[' && p[1] == ']')
	{
		if(*flags & TYPFLAG_POINTER)
		{
			PQseterror(
				"'*' specifer flag cannot be used with arrays[] '%s' (pos:%d)",
				id, typpos);
			return NULL;
		}

		*flags |= TYPFLAG_ARRAY;
		p += 2;
	}

	return p;
}


/* Does an OID lookup based only on typname, no typnamespace provided.
 * Takes the search path into account.
 */
#define OIDLOOKUP \
"SELECT t.oid AS typoid, a.oid AS arroid, t.typlen FROM pg_type t" \
"  JOIN pg_type a ON ('_' || t.typname, t.typnamespace) = (a.typname, a.typnamespace)" \
"  JOIN pg_namespace n ON t.typnamespace = n.oid" \
"  JOIN" \
"  (" \
"    SELECT * FROM information_schema._pg_expandarray(current_schemas(true))" \
"  ) s ON s.x = n.nspname" \
"  WHERE t.typname = $1 ORDER BY s.n LIMIT 1"

/* Does an OID lookup based on typname and typnamespace. Since the
 * typnamespace is provided, there is no need to use the search path.
 */
#define SCHEMA_OIDLOOKUP \
"SELECT t.oid AS typoid, a.oid AS arroid, t.typlen FROM pg_type t" \
"  JOIN pg_type a ON ('_' || t.typname, t.typnamespace) = (a.typname, a.typnamespace)" \
"  JOIN pg_namespace n ON t.typnamespace = n.oid" \
"  WHERE n.nspname = $1 AND t.typname = $2"

/* record attribute lookup by OID: oid, name, len, typmod */
#define RECATTLOOKUP \
"SELECT" \
"  CASE when t.typtype = 'd' then t.typbasetype else a.atttypid END as typoid," \
"  a.attname, a.attlen, a.atttypmod " \
"FROM pg_type b" \
"  JOIN pg_attribute a ON b.typrelid = a.attrelid" \
"  JOIN pg_type t ON a.atttypid = t.oid" \
"  WHERE b.oid = $1" \
"    AND a.attnum > 0" \
"    AND NOT a.attisdropped" \
"  ORDER BY a.attnum"

static int
expandHandlers(PGtypeData *typeData)
{
	int hmax;
	PGtypeHandler *h;

	if(typeData->typhcnt < typeData->typhmax)
		return TRUE;

	hmax = typeData->typhmax ? (typeData->typhmax * 3) / 2 : 8;
	h = (PGtypeHandler *)pqt_realloc(
		typeData->typhandlers, sizeof(PGtypeHandler) * hmax);

	if(!h)
	{
		PQseterror(PQT_OUTOFMEMORY);
		return FALSE;
	}

	typeData->typhandlers = h;
	typeData->typhmax = hmax;
	return TRUE;
}

static int
createHandler(PGconn *conn, PGtypeData *connData, PGtypeHandler *handler,
	char *typschema, char *typname, PGtypeProc typput, PGtypeProc typget)
{
	PGresult *res;
	PGint2 typlen;
	char fqtn[200];
	PGparam *param = PQparamCreate(conn);
	const char *command = OIDLOOKUP;

	if(!param)
		return -1;

	if(typschema && *typschema)
	{
		if(!PQputf(param, "%name*", typschema))
		{
			PQparamClear(param);
			return -1;
		}

		command = SCHEMA_OIDLOOKUP;
	}

	if(!PQputf(param, "%name*", typname))
	{
		PQparamClear(param);
		return -1;
	}

	res = PQparamExec(conn, param, command, 1);
	PQparamClear(param);
	if(!res)
		return -1;

	if(PQresultStatus(res) != PGRES_TUPLES_OK)
	{
		PQclear(res);
		return -1;
	}

	if(PQntuples(res) == 0)
	{
		PQseterror(
			"server type lookup failed: could not find '%s'",
			pqt_fqtn(fqtn, sizeof(fqtn), typschema, typname));
		PQclear(res);
		return -1;
	}

	memset(handler, 0, sizeof(PGtypeHandler));

	if(!PQgetf(res, 0, "%oid %oid %int2",
			0, &handler->typoid,
			1, &handler->typoid_array,
			2, &typlen))
	{
		PQclear(res);
		return -1;
	}

	PQclear(res);

	handler->id        = connData->typhcnt + DEFAULT_HANDLER_COUNT;
	handler->typschema = (typschema && *typschema) ? strdup(typschema) : NULL;
	handler->typname   = strdup(typname);
	handler->typlen    = (int)typlen;
	handler->typput    = typput;
	handler->typget    = typget;
	handler->base_id   = -1;

	return 0;
}

static int
registerSubClass(PGtypeData *connData, const char *basestr,
	char *sub_typschema, char *sub_typname, const char *orig_typname,
	PGtypeProc typput, PGtypeProc typget)
{
	int flags;
	PGtypeHandler *sub_handler;
	PGtypeHandler *base_handler;
	char base_typschema[PQT_MAXIDLEN + 1];
	char base_typname[PQT_MAXIDLEN + 1];
	char sub_fqtn[200];
	char base_fqtn[200];

	if(!pqt_parsetype(basestr, base_typschema, base_typname, &flags, 1))
		return FALSE;

	if(flags & TYPFLAG_INVALID)
		return FALSE;

	if(flags & TYPFLAG_ARRAY)
	{
		PQseterror("Cannot use an array[] during type registration.");
		return FALSE;
	}

	if(flags & TYPFLAG_POINTER)
	{
		PQseterror("Cannot use a type* during type registration.");
		return FALSE;
	}

	/* lookup the base handler */
	base_handler = pqt_gethandler(
		connData->typhandlers,
		connData->typhcnt,
		base_typschema,
		base_typname);

	if(!base_handler)
	{
		PQseterror(
			"typname '%s' does not exist, '%s' cannot sub-class it",
			pqt_fqtn(base_fqtn, sizeof(base_fqtn), base_typschema, base_typname),
			pqt_fqtn(sub_fqtn, sizeof(sub_fqtn), sub_typschema, sub_typname));
		return FALSE;
	}

	/* cannot sub-class a record type */
	if(base_handler->typoid == RECORDOID)
	{
		PQseterror("Cannot sub-class pg_catalog.record '%s'",
			pqt_fqtn(sub_fqtn, sizeof(sub_fqtn), sub_typschema, sub_typname));
		return FALSE;
	}

	if(!expandHandlers(connData))
		return FALSE;

	sub_handler               = &connData->typhandlers[connData->typhcnt];
	sub_handler->id           = connData->typhcnt + DEFAULT_HANDLER_COUNT;
	sub_handler->typschema    = (sub_typschema && *sub_typschema) ?
                                strdup(sub_typschema) : NULL;
	sub_handler->typname      = strdup(sub_typname);
	sub_handler->typlen       = base_handler->typlen;
	sub_handler->typoid       = base_handler->typoid;
	sub_handler->typoid_array = base_handler->typoid_array;
	sub_handler->typput       = typput;
	sub_handler->typget       = typget;
	sub_handler->base_id      = base_handler->id;
	sub_handler->nattrs       = 0;
	sub_handler->attDescs     = NULL;
	sub_handler->orig_typname = strdup(orig_typname);
	sub_handler->orig_typput  = typput;
	sub_handler->orig_typget  = typget;
	connData->typhcnt++;
	return TRUE;
}

#define COMPCLEAR do{ \
	if(comp_handler->typschema) free(comp_handler->typschema); \
	if(comp_handler->typname) free(comp_handler->typname); \
} while(0)

static int
registerRecord(PGconn *conn, PGtypeData *connData,
	char *comp_typschema, char *comp_typname, const char *orig_typname)
{
	int tup;
	int ntups;
	PGresult *res;
	PGtext attname;
	PGint2 attlen;
	char fqtn[200];
	PGrecordAttDesc *attDescs;
	PGtypeHandler *rec_handler;
	PGtypeHandler *comp_handler;
	PGparam *param;

	/* cannot fail */
	rec_handler = pqt_gethandler(NULL, 0, "pg_catalog", "record");

	/* make sure conn's type handlers array is large enough */
	if(!expandHandlers(connData))
		return FALSE;

	/* create the handler */
	comp_handler = &connData->typhandlers[connData->typhcnt];
	if(createHandler(
			conn,
			connData,
			comp_handler,
			comp_typschema,
			comp_typname,
			rec_handler->typput,
			rec_handler->typget) == -1)
		return FALSE;

	/* Invalid, you cannot register a record composite */
	if(comp_handler->typoid == rec_handler->typoid)
	{
		COMPCLEAR;
		PQseterror(
			"record already exists, its a built-in type");
		return FALSE;
	}

	param = PQparamCreate(conn);
	if(!param)
	{
		COMPCLEAR;
		return FALSE;
	}

	if(!PQputf(param, "%oid", comp_handler->typoid))
	{
		COMPCLEAR;
		PQparamClear(param);
		return FALSE;
	}

	res = PQparamExec(conn, param, RECATTLOOKUP, 1);
	PQparamClear(param);

	if(!res)
	{
		COMPCLEAR;
		return FALSE;
	}

	if(PQresultStatus(res) != PGRES_TUPLES_OK)
	{
		COMPCLEAR;
		PQclear(res);
		return FALSE;
	}

	if((ntups = PQntuples(res)) == 0)
	{
		COMPCLEAR;
		PQseterror(
			"Failed looking up composite attributes: Is '%s' a composite?",
			pqt_fqtn(fqtn, sizeof(fqtn), comp_handler->typschema,
			comp_handler->typname));
		PQclear(res);
		return FALSE;
	}

	attDescs = (PGrecordAttDesc *)malloc(ntups * sizeof(PGrecordAttDesc));
	if(!attDescs)
	{
		COMPCLEAR;
		PQseterror(PQT_OUTOFMEMORY);
		PQclear(res);
		return FALSE;
	}

	for(tup=0; tup < ntups; tup++)
	{
		/* composite field attributes: attoid, attname, attlen, atttypmod */
		if(!PQgetf(res, tup, "%oid %name %int2 %int4",
				0, &attDescs[tup].attoid,
				1, &attname,
				2, &attlen,
				3, &attDescs[tup].atttypmod))
		{
			COMPCLEAR;
			PQclear(res);
			return FALSE;
		}

		attDescs[tup].attlen = (int)attlen;
		attDescs[tup].attname = strdup(attname);
		if(!attDescs[tup].attname)
		{
			COMPCLEAR;
			PQclear(res);
			return FALSE;
		}
	}

	PQclear(res);

	comp_handler->orig_typname = strdup(orig_typname);
	comp_handler->nattrs = ntups;
	comp_handler->attDescs = attDescs;
	connData->typhcnt++;
	return TRUE;
}




