
/*
 * hooks.c
 *   The libpq Object Hooks implementation.
 *
 * Copyright (c) 2008 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

/* Hook Functions */
static void *hook_initHookData(const PGconn *conn);
static void  hook_connReset(const PGconn *conn);
static void  hook_connDestroy(const PGconn *conn);
static void *hook_resultCreate(const PGconn *conn, const PGresult *res);
static void *hook_resultCopy(PGresult *dest, const PGresult *src);
static void  hook_resultDestroy(const PGresult *res);

/* helper functions */
static PGtypeData *allocTypeData(void);
static void freeTypeData(PGtypeData *typeData);

PGobjectHooks *
PQtypesObjectHooks(void)
{
	static PGobjectHooks pqt_hooks = {
		PQT_OBJHOOKSNAME,
		NULL, /* data */
		hook_initHookData,
		hook_connReset,
		hook_connDestroy,
		hook_resultCreate,
		hook_resultCopy,
		hook_resultDestroy
	};

	return &pqt_hooks;
}

static void *
hook_initHookData(const PGconn *conn)
{
	return allocTypeData();
}

static void
hook_connReset(const PGconn *conn)
{
	int i;
	int hcnt;
	PGtypeHandler *handlers;
	PGtypeData *connData = (PGtypeData *)PQhookData(conn, PQT_OBJHOOKSNAME);

	if(!connData)
		return;

	/* Have to re-register the handlers to help avoid invalidation issues.
	 * Save off the handlers and set the conn's type handlers to NULL.
	 */
	hcnt = connData->typhcnt;
	handlers = connData->typhandlers;
	connData->typhcnt = 0;
	connData->typhandlers = NULL;

	/* re-register type handlers.  It is entirely possible that a register
	 * will fail. This would almost always be because a previously registered
	 * type no longer exists in the backend (corner case).  We silently
	 * ignore these and let putf/getf alert the user ... "no such type".
	 */
	for(i=0; i < hcnt; i++)
		PQregisterTypeHandler((PGconn *)conn, handlers[i].orig_typname,
			handlers[i].orig_typput, handlers[i].orig_typget);

	/* free the out-dated handler list */
	pqt_freehandlers(handlers, hcnt);
}

static void
hook_connDestroy(const PGconn *conn)
{
	freeTypeData((PGtypeData *)PQhookData(conn, PQT_OBJHOOKSNAME));
}

static void *
hook_resultCreate(const PGconn *conn, const PGresult *res)
{
	PGtypeData *resData;
	PGtypeData *connData = (PGtypeData *)PQhookData(conn, PQT_OBJHOOKSNAME);

	if(!connData || !(resData = allocTypeData()))
		return NULL;

	/* get type formatting info from conn */
	pqt_getfmtinfo(conn, &resData->fmtinfo);

	/* copy type handlers from PGconn's typeData */
	if(connData->typhcnt > 0)
	{
		resData->typhandlers = pqt_duphandlers(
			connData->typhandlers, connData->typhcnt);

		if(resData->typhandlers)
			resData->typhcnt = connData->typhcnt;
	}

	return resData;
}

static void *
hook_resultCopy(PGresult *dest, const PGresult *src)
{
	PGtypeData *destData;
	PGtypeData *srcData = (PGtypeData *)PQresultHookData(src, PQT_OBJHOOKSNAME);

	if(!srcData || !(destData = allocTypeData()))
		return NULL;

	memcpy(&destData->fmtinfo, &srcData->fmtinfo, sizeof(PGtypeFormatInfo));

	/* copy type handlers from PGresult's typeData */
	if(srcData->typhcnt > 0)
	{
		destData->typhandlers = pqt_duphandlers(
			srcData->typhandlers, srcData->typhcnt);

		if(destData->typhandlers)
			destData->typhcnt = srcData->typhcnt;
	}

	return destData;
}

static void
hook_resultDestroy(const PGresult *res)
{
	freeTypeData((PGtypeData *)PQresultHookData(res, PQT_OBJHOOKSNAME));
}

static PGtypeData *
allocTypeData(void)
{
	PGtypeData *typeData = (PGtypeData *)malloc(sizeof(PGtypeData));

	if(typeData)
		memset(typeData, 0, sizeof(PGtypeData));

	return typeData;
}

static void
freeTypeData(PGtypeData *typeData)
{
	if(typeData)
	{
		pqt_freehandlers(typeData->typhandlers, typeData->typhcnt);
		free(typeData);
	}
}

